open Useful

let mutex = Mutex.create ()

let draw_segment (x, y) (x', y') =
(*  Mutex.lock mutex;*)
  Graphics.moveto x y;
  Graphics.lineto x' y'(*;
  Mutex.unlock mutex*)

let autoplay_quite = ref true

open Iv2

type shape_type = Square | Hollow_square | Hollow_circle | Circle | Smarty

type border_type = No | Dots | Sides | Lines

type state_type =
  { shift:int * int;
    scale:int;
    cell_colour_nbr:int;
    cell_colour:int;
    bg_colour:int;
    shape:shape_type;
    borders:border_type }

let default_cell_colour = 10
 
let default =
  { shift = zero_zero;
    scale = 32;
    cell_colour_nbr = default_cell_colour;
    cell_colour = Colour.colour_circle default_cell_colour;
    bg_colour = Graphics.black;
    shape = Square;
    borders = Lines }

let config_name = "funLife.config" 

let state =
  ref
    (try
      let chan = open_in_bin config_name in
      let saved_config = (Marshal.from_channel chan : state_type) in
        close_in chan;
        saved_config
    with Sys_error "funLife.config: No such file or directory" ->
      default)

let save_state () =
  let chan = open_out_bin config_name in
    Marshal.to_channel chan !state [];
    close_out chan

let images =
  ref (Io.load_images (if !state.scale = 1 then 1 else !state.scale-1))

let clear () =
  Graphics.set_color !state.bg_colour;
  Graphics.fill_rect 0 0 (Graphics.size_x () - 1) (Graphics.size_y () - 1)

let  _ =
  Graphics.open_graph (Sys.getenv "DISPLAY" ^ " 800x600");
    Graphics.set_window_title "Fun Life";
    Graphics.display_mode false;
    clear ();
    Graphics.synchronize ()

let close = Graphics.close_graph

let phys_size = Graphics.size_x (), Graphics.size_y ()

let set_state s = state := s

let center = phys_size // 2

let delta = ref zero_zero

let update_delta () = delta := center ++ (!state.scale/2) ** (1, 1)

let _ = update_delta ()

let physical_of_logical c =
  let st = !state in
    st.scale ** (c -- st.shift) ++ !delta

let logical_of_physical c =
  let st = !state in
    (div2 (c -- !delta) st.scale) ++ st.shift

let colourised_cell c =
  if Iv2Set.mem c (WorldHistory.get_next ()) then
    if Iv2Set.mem c (WorldHistory.get_prev ()) then
      Colour.blue
    else
      Colour.green
  else
    if Iv2Set.mem c (WorldHistory.get_prev ()) then
      Colour.red
    else
      Colour.yellow

let ls0 = ref zero_zero
let lsn = ref zero_zero
      
let update_logical_screen () =
  let l0, ln =
    Couple.map logical_of_physical (zero_zero, phys_size -- (1, 1)) in
      ls0 := l0; lsn := ln

let _ = update_logical_screen ()

open Couple

let draw_dot c =
  if !ls0 <<<= c && c <<<= !lsn then (
    let v = !state in
    let x, y = physical_of_logical c in
    let side = pred v.scale in
    (*    Mutex.lock mutex;*)
      Graphics.set_color v.cell_colour;
      (match v.shape with
        Square ->
          Graphics.fill_rect x y side side;
      | Hollow_square ->
          Graphics.fill_rect x y side side;
          Graphics.set_color !state.bg_colour;
          let half = v.scale/2 in
          let quarter = half/2 in
          Graphics.fill_rect (x+quarter) (y+quarter) (half-1) (half-1);
      | Hollow_circle ->
          let r = v.scale/2 in
            Graphics.fill_circle (x+r) (y+r) r;
            Graphics.set_color !state.bg_colour;
            Graphics.fill_circle (x+r) (y+r) (r/2);
      | Circle ->
          let r = v.scale/2 in
            Graphics.fill_circle (x+r) (y+r) r;
      | Smarty ->
          Graphic_image.draw_image
            !images.(colourised_cell c)
            (x+1) (y+1));
      if v.shape = Square || v.shape = Hollow_square then(
        Graphics.set_color Colour.grey;
        match v.borders with
          No | Lines ->
            ()
        | Dots ->
          Graphics.plot x y
        | Sides ->
          Graphics.moveto (x+side) y;
          Graphics.lineto x y;
          Graphics.lineto x (y+side))(*;
      Mutex.unlock mutex*)
  )

let draw_inside_square c =
  let x, y = physical_of_logical c in
    let half = !state.scale/2 in
    let quarter = half/2 in
      Graphics.fill_rect (x+quarter) (y+quarter) (half-1) (half-1)

let draw_inside_circle c =
  let x, y = physical_of_logical c in
  let r = !state.scale/2 in
    Graphics.fill_circle (x+r) (y+r) (r/2)

let shift d = set_state (let v = !state in { v with shift = v.shift ++ d })

let scale op =
  set_state (
    let v = !state in
    let new_scale = op v.scale 2 in
      if new_scale = 0 || new_scale = 2*sqr default.scale then
        v
      else(
        if new_scale <> 1 then
          images := Io.load_images (pred new_scale);
        { v with scale = new_scale }))

let next_bg c =
  if c = Graphics.black then
    Graphics.white
  else
    Graphics.black

let mouse_colour () =
  Mutex.lock mutex;
  let c =
    try
    
      let check xy =
        if zero_zero <<<= xy && xy <<< phys_size then
          xy
        else
          raise Exit in
        
        Some 
          (Couple.uncurry2
            Graphics.point_color
            (check (Graphics.mouse_pos ())))
    with
      Exit -> None in
  Mutex.unlock mutex;
  c

let change_colour () =
  set_state(
    let v = !state in
    match mouse_colour () with
    Some mc ->
      if mc = v.cell_colour then(
        let v' = { v with cell_colour_nbr = Colour.next v.cell_colour_nbr } in
          { v' with cell_colour = Colour.colour_circle v'.cell_colour_nbr }
      )else
        if mc = v.bg_colour then
          { v with bg_colour = next_bg v.bg_colour }
        else
          v
  | None ->
    v)

let change_shape () =
  set_state(
    let v = !state in
    { v with shape =
        match v.shape with
          Square ->
            Hollow_square
        | Hollow_square ->
            Hollow_circle
        | Hollow_circle ->
            Circle
        | Circle ->
            Smarty
        | Smarty ->
            Square })

let change_borders v =
  set_state(
    let v = !state in
    { v with borders =
        match v.borders with
          No ->
            Dots
        | Dots ->
            Sides
        | Sides ->
           Lines
        | Lines ->
            No })

let draw () =
  update_logical_screen ();
  Mutex.lock mutex;
  Graphics.display_mode false;
  clear ();
  update_delta ();
  let v = !state in
    if v.scale = 1 then(
      let delta2 = !delta -- v.shift in
        if v.shape = Smarty then
          Iv2Set.iter
            (fun c ->
               Graphics.set_color Colour.colours.(colourised_cell c);
               (Couple.uncurry2 Graphics.plot <<- ((++) delta2)) c)
            (WorldHistory.get_state ())
        else(
          Graphics.set_color v.cell_colour;
            Iv2Set.iter
              (Couple.uncurry2 Graphics.plot <<- ((++) delta2))
              (WorldHistory.get_state ())
        )
    )else(    
    (*  Mutex.unlock mutex;*)
      Iv2Set.iter draw_dot (WorldHistory.get_state ());
      if !state.shape == Hollow_square then(
        Graphics.set_color v.cell_colour;
        Iv2Set.iter draw_inside_square (WorldHistory.get_next ())
      )else
        if !state.shape == Hollow_circle then(
          Graphics.set_color v.cell_colour;
          Iv2Set.iter draw_inside_circle (WorldHistory.get_next ())
        );
    let side = v.scale in
      if side > 1 && v.borders = Lines then(
        Graphics.set_color Colour.grey;
        let (x0, y0) = physical_of_logical !ls0 in
        let (xn, yn) = phys_size // side ++ (1,1) in
        let size_x, size_y = phys_size in
          for i = 0 to xn do
            let x = x0+i*side in
              draw_segment (x, 0) (x, size_y-1)
          done;
          for i = 0 to yn do
            let y = y0+i*side in
              draw_segment (0, y) (size_x-1, y)
          done));
  (*  Mutex.lock mutex;*)
  Graphics.synchronize ();
  Mutex.unlock mutex

let random_screen world =
  let world' = ref world in
    update_logical_screen ();
    let lsx0, lsy0 = !ls0 in
    let lsxn, lsyn = !lsn in
      for x = lsx0 to lsxn do
        for y = lsy0 to lsyn do
          if Random.bool () then
            world' := Iv2Set.add (x, y) !world'
        done
      done;
      !world'

let random () =
  WorldHistory.change_state random_screen

let rec drag org rem nothing_done =
  let status =
    Graphics.wait_next_event [Graphics.Button_up; Graphics.Mouse_motion] in
  let now = status.Graphics.mouse_x, status.Graphics.mouse_y in
  let physical_delta = rem ++ now -- org in
  let v = !state in
  let logical_delta = physical_delta // v.scale in
  let detected_move = logical_delta <> zero_zero in
  let nothing_done_here = nothing_done && not detected_move in
    set_state { v with shift = v.shift -- logical_delta };
    if !autoplay_quite && detected_move then
      draw ();
    if status.Graphics.button then
      drag
        now
        (physical_delta -- v.scale ** logical_delta)
        nothing_done_here
    else
      nothing_done_here

let wait_action key mouse =
  let status =
    Graphics.wait_next_event [Graphics.Key_pressed; Graphics.Button_down] in
    (if status.Graphics.keypressed then
      key status.Graphics.key
    else (* Button_down *)
      let mouse_org = status.Graphics.mouse_x, status.Graphics.mouse_y in
        if drag mouse_org zero_zero true then
          mouse (logical_of_physical mouse_org));
    draw ()

let text_page s =
  Printf.printf "%s" s;
  flush stdout